import matplotlib.pyplot as plt
import numpy as np
import torch
from sklearn.metrics import r2_score  # type: ignore

import reptrix.utils as utils


def get_powerlaw(eigen: np.ndarray, trange: np.ndarray) -> tuple:
    """Fit powerlaw and return decay, powerlaw fit and the goodness of fit

    Args:
        eigen (np.ndarray): Eigenspectrum of activation covariance matrix
        trange (np.ndarray): Range to fit the powerlaw.
                            Tip: Ignore the first couple of eigenvalues
                                because we want to fit the tail of the spectrum

    Returns:
        tuple: Result of size 4
            alpha: powerlaw decay coefficient
            ypred: Powerlaw fit
            fit_R2: goodness of powerlaw fit
            fit_R2_100: goodness of powerlaw fit (computed till the 100 eigvals)
    """
    # Inspired by Stringer+Pachitariu 2018b github repo!
    # (https://github.com/MouseLand/stringer-pachitariu-et-al-2018b/blob/master/python/utils.py)
    logss = np.log(np.abs(eigen))
    y = logss[trange][:, np.newaxis]
    trange += 1
    nt = trange.size
    x = np.concatenate((-np.log(trange)[:, np.newaxis], np.ones((nt, 1))), axis=1)
    w = 1.0 / trange.astype(np.float32)[:, np.newaxis]
    b = np.linalg.solve(x.T @ (x * w), (w * x).T @ y).flatten()

    allrange = np.arange(0, eigen.size).astype(int) + 1
    x = np.concatenate(
        (-np.log(allrange)[:, np.newaxis], np.ones((eigen.size, 1))), axis=1
    )
    ypred = np.exp((x * b).sum(axis=1))
    alpha = b[0]
    max_range = (
        500 if len(eigen) >= 512 else len(eigen) - 10
    )  # subtracting 10 here arbitrarily because we want to avoid the last tail!
    fit_R2 = r2_score(
        y_true=logss[trange[0] : max_range],
        y_pred=np.log(np.abs(ypred))[trange[0] : max_range],
    )
    try:
        fit_R2_100 = r2_score(
            y_true=logss[trange[0] : 100], y_pred=np.log(np.abs(ypred))[trange[0] : 100]
        )
    except:  # noqa: E722
        fit_R2_100 = None
    return (alpha, ypred, fit_R2, fit_R2_100)


def get_alpha(
    activations: torch.Tensor,
    max_eigenvals: int = 2048,
    fit_range: np.ndarray = None,
) -> tuple:
    """Get alpha and powerlaw fit
    (https://proceedings.neurips.cc/paper_files/paper/2022/hash/70596d70542c51c8d9b4e423f4bf2736-Abstract-Conference.html)

    Args:
        activations (np.ndarray): Activation tensor of shape (bsz,d1,d2...dn)
        max_eigenvals (int, optional): Maximum #eigenvalues to compute.
                                    Defaults to 2048.
        fit_range (np.ndarray, optional): Range to fit the powerlaw.
                                        Defaults to np.arange(10,100).

    Returns:
        tuple: Result of size 4
            alpha: powerlaw decay coefficient
            ypred: Powerlaw fit
            fit_R2: goodness of powerlaw fit
            fit_R2_100: goodness of powerlaw fit (computed till the 100 eigvals)
    """
    if fit_range is None:
        fit_range = np.arange(10, 100)
    if activations.requires_grad:
        activations_arr = activations.detach().cpu()
    else:
        activations_arr = activations.cpu()
    activations_arr_np = activations_arr.numpy()
    eigen = utils.get_eigenspectrum(
        activations_np=activations_arr_np, max_eigenvals=max_eigenvals
    )
    alpha_res = get_powerlaw(eigen=eigen, trange=fit_range)
    return alpha_res


def plot_powerlaw(eigenspectrum: np.ndarray, alpha_res: tuple) -> None:
    """Plot eigenspectrum and powerlaw fit

    Args:
        eigenspectrum (np.ndarray): Eigenspectrum of activation covariance matrix
        alpha_res (tuple): Tuple containing alpha, powerlaw fit,
                        goodness of powerlaw fit,
                        goodness of powerlaw fit for first 100 eigenvalues
    """
    utils.plot_eigenspectrum(eigenspectrum=eigenspectrum)
    alpha, ypred, r2, r2_100 = alpha_res
    xrange = np.arange(1, 1 + len(ypred))
    plt.loglog(xrange, ypred, c='red', ls='--', lw=2.0, label='Powerlaw fit')
    plt.title(
        rf'Eigenspectrum and powerlaw fit, $\alpha$ = {alpha:.3f} '
        + rf'(r2_100 = {r2_100:.3f})'
    )



import numpy as np
import torch

from reptrix import alpha


def test_get_alpha() -> None:
    np.random.seed(0)
    torch.manual_seed(0)
    activations_arr = torch.randn(1000, 1024)

    metric_alpha = alpha.get_alpha(activations_arr)
    # print(metric_alpha[0], metric_alpha[2], metric_alpha[3])
    assert np.allclose(metric_alpha[0], 0.148936)



import numpy as np
import torch

from reptrix import lidar


def test_get_lidar() -> None:
    np.random.seed(0)
    torch.manual_seed(0)
    activations_arr = torch.randn(1000, 10, 1024)

    metric_lidar = lidar.get_lidar(activations_arr, 1000, 10)
    assert np.allclose(metric_lidar, 348.08795)



import numpy as np
import torch

import reptrix.utils as utils


def get_rank(eigen: np.ndarray) -> float:
    """Get effective rank of the representation covariance matrix

    Args:
        eigen (np.ndarray): Eigenspectrum of the representation covariance matrix

    Returns:
        float: Effective rank
    """
    l1 = np.sum(np.abs(eigen))
    eps = 1e-7
    eigen_norm = eigen / l1 + eps
    entropy = -np.sum(eigen_norm * np.log(eigen_norm))
    return np.exp(entropy)


def get_rankme(activations: torch.Tensor, max_eigenvals: int = 2048) -> float:
    """Get RankMe metric
    (https://proceedings.mlr.press/v202/garrido23a)

    Args:
        activations (np.ndarray): Activation tensor of shape (bsz,d1,d2...dn)
        max_eigenvals (int, optional): Maximum #eigenvalues to compute.
                                    Defaults to 2048.

    Returns:
        float: RankMe metric
    """
    if activations.requires_grad:
        activations_arr = activations.detach().cpu()
    else:
        activations_arr = activations.cpu()
    activations_arr_np = activations_arr.numpy()
    eigen = utils.get_eigenspectrum(
        activations_np=activations_arr_np, max_eigenvals=max_eigenvals
    )
    rank = get_rank(eigen)
    return rank



import numpy as np
import torch

import reptrix.utils as utils


def get_rank(eigen: np.ndarray) -> float:
    """Get effective rank of the LDA covariance matrix

    Args:
        eigen (np.ndarray): Eigenspectrum of the LDA covariance matrix

    Returns:
        float: Effective rank
    """
    l1 = np.sum(np.abs(eigen))
    eps = 1e-7
    eigen_norm = eigen / l1 + eps
    entropy = -np.sum(eigen_norm * np.log(eigen_norm))
    return np.exp(entropy)


def get_lidar(
    activations: torch.Tensor,
    num_samples: int,
    num_augs: int,
    del_sigma_augs: float = 1e-6,
    max_eigenvals: int = 2048,
) -> float:
    """Get RankMe metric
    (https://openreview.net/forum?id=f3g5XpL9Kb)

    Args:
        activations (torch.Tensor): Activation tensor of shape either
                                    (num_samples, num_augs, d1,d2...dn) or
                                    (num_samples * num_augs, d1,d2...dn)
        num_samples (int): Number of unique inputs/samples
        num_augs (int): Number of augmentations for each input used to estimate
                        object manifold
        del_sigma_augs (float, optional): A small positive constant that is added
                                        to make sure matrices are invertible.
                                        Defaults to 1e-6.
        max_eigenvals (int, optional): Maximum #eigenvalues to compute.
                                    Defaults to 2048.

    Returns:
        float: LiDAR metric
    """
    if activations.requires_grad:
        activations_arr = activations.detach().cpu()
    else:
        activations_arr = activations.cpu()
    if activations_arr.shape[0] == num_samples * num_augs:
        activations_arr = activations_arr.reshape(num_samples, num_augs, -1)
    else:
        d0 = activations_arr.shape[0]
        d1 = activations_arr.shape[1]
        assert (
            d0 == num_samples and d1 == num_augs
        ), "Tensor activations should have shape (num_samples, num_augs,...)"
        activations_arr = activations_arr.reshape(d0, d1, -1)

    object_activations = activations_arr.mean(dim=1, keepdim=True)  # mean over augs
    mean_activations = object_activations.mean(dim=0, keepdim=True)
    # compute the inter-object manifold covariance
    sigma_obj = (object_activations - mean_activations).squeeze().T @ (
        object_activations - mean_activations
    ).squeeze()

    # compute the intra-object manifold covariance and take mean across objects
    sigma_augs = torch.bmm(
        (activations_arr - object_activations).permute((0, 2, 1)),
        (activations_arr - object_activations),
    ).mean(dim=0)
    # add Identity to ensure invertibility
    sigma_augs += del_sigma_augs * torch.eye(sigma_augs.shape[0])
    sigma_augs_inv_sqrt = utils.mat_sqrt_inv(sigma_augs)

    # compute LIDAR matrix
    sigma_lidar = sigma_augs_inv_sqrt @ sigma_obj @ sigma_augs_inv_sqrt
    sigma_lidar_np = sigma_lidar.numpy()

    eigen = utils.get_eigenspectrum(
        activations_np=sigma_lidar_np, max_eigenvals=max_eigenvals
    )
    lidar = get_rank(eigen)
    return lidar
